# standard tags (binary annotations)
HTTP_HOST = "http.host"
HTTP_METHOD = "http.method"
HTTP_PATH = "http.path"
HTTP_REQUEST_SIZE = "http.request.size"
HTTP_RESPONSE_SIZE = "http.response.size"
HTTP_STATUS_CODE = "http.status_code"
HTTP_URL = "http.url"
HTTP_ROUTE = "http.route"

ERROR = "error"
LOCAL_COMPONENT = "lc"

CLIENT_ADDR = "ca"
MESSAGE_ADDR = "ma"
SERVER_ADDR = "sa"
